/* db.c */

/*
 * Copyright (c) 1997-2002 Hoshi Takanori
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <PalmOS.h>

#include "life.h"


/*
 * constants
 */

#define noErr 0
#define DEFAULT_CARD_NO 0

#define LIFE_DB_TYPE 'DATA'
#define LIFE_DB_NAME "Pocket Life Data"

#define LIFE_DATA_INDEX 0


/*
 * types
 */

struct header {
    offset_t init_offset;
    offset_t init_length;
    gen_t generation;
    offset_t cur_length;
};


/*
 * functions
 */

void read_db(cell_t *tape, offset_t tape_length,
    offset_t *init_length, gen_t *generation, offset_t *cur_length)
{
    DmOpenRef db;
    MemHandle h;
    long size;
    const struct header *p;
    offset_t offset, length;

    *init_length = 0;
    *generation = 0;
    *cur_length = 0;

    db = DmOpenDatabaseByTypeCreator(LIFE_DB_TYPE, CREATOR, dmModeReadOnly);
    if (db != NULL) {
	h = DmQueryRecord(db, LIFE_DATA_INDEX);
	if (h != NULL) {
	    size = MemHandleSize(h);
	    if (size > sizeof(struct header)) {
		p = MemHandleLock(h);

		offset = p->init_offset;
		length = p->init_length;
		if ((offset & 1) == 0
		 && offset >= sizeof(struct header) && offset < size
		 && length > 0 && length < tape_length / 2 - 1
		 && offset + sizeof(cell_t) * length <= size) {
		    MemMove(tape, (char *) p + offset, sizeof(cell_t) * length);
		    *init_length = length;

		    offset += sizeof(cell_t) * length;
		    length = p->cur_length;
		    if (p->generation > 0
		     && length > 0 && length < tape_length - *init_length - 1
		     && offset + sizeof(cell_t) * length <= size) {
			MemMove(tape + *init_length + 1,
				(char *) p + offset, sizeof(cell_t) * length);
			*generation = p->generation;
			*cur_length = length;
		    }
		}

		MemHandleUnlock(h);
	    }
	}
	DmCloseDatabase(db);
    }
}

void write_db(const cell_t *tape, offset_t init_length,
    gen_t generation, const cell_t *life_begin, const cell_t *life_end)
{
    DmOpenRef db;
    Err err;
    struct header header;
    long size, offset, length1, length2;
    UInt16 index;
    MemHandle h;
    Char *p;

    db = DmOpenDatabaseByTypeCreator(LIFE_DB_TYPE, CREATOR, dmModeReadWrite);
    if (db == NULL) {
	err = DmCreateDatabase(
		DEFAULT_CARD_NO, LIFE_DB_NAME, CREATOR, LIFE_DB_TYPE, false);
	if (err == noErr) {
	    db = DmOpenDatabaseByTypeCreator(
		    LIFE_DB_TYPE, CREATOR, dmModeReadWrite);
	}
    }

    if (db != NULL) {
	header.init_offset = sizeof header;
	header.init_length = init_length;
	header.generation = generation;
	header.cur_length =
		(generation == 0) ? 0 : count_length(life_begin, life_end);
	size = sizeof header
		+ sizeof(cell_t) * (init_length + header.cur_length);

	index = LIFE_DATA_INDEX;
	if (DmNumRecords(db) > 0) {
	    h = DmResizeRecord(db, index, size);
	} else {
	    h = DmNewRecord(db, &index, size);
	}

	if (h != NULL) {
	    p = MemHandleLock(h);
	    DmWrite(p, 0, &header, sizeof header);
	    DmWrite(p, sizeof header, tape, sizeof(cell_t) * init_length);
	    if (header.cur_length > 0) {
		offset = sizeof header + sizeof(cell_t) * init_length;
		if (life_begin <= life_end) {
		    DmWrite(p, offset,
			    life_begin, sizeof(cell_t) * header.cur_length);
		} else {
		    length2 = life_end - (tape + init_length);
		    length1 = header.cur_length - length2;
		    DmWrite(p, offset, life_begin, sizeof(cell_t) * length1);
		    DmWrite(p, offset + sizeof(cell_t) * length1,
			    life_end - length2, sizeof(cell_t) * length2);
		}
	    }
	    MemHandleUnlock(h);

	    DmReleaseRecord(db, index, true);
	}

	DmCloseDatabase(db);
    }
}
